import enum
import re
import warnings
from collections.abc import Mapping
from typing import Optional, Union

from servicefoundry.auto_gen import models
from servicefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    constr,
    root_validator,
    validator,
)

LEGACY_GPU_TYPE_COUNT_WARNING_MESSAGE_TEMPLATE = """
---------
The `gpu_count` and `gpu_type` fields are deprecated. Please remove these fields
from your deployment Python script or YAML Spec.

If you are using Python SDK, add GPUs in the following way:

```
from servicefoundry import NvidiaGPU, Resources
...

  resources=Resources(
    ...
    devices=[NvidiaGPU(name="{gpu_type}", count={gpu_count})],
  )
```

If you are using YAML Spec to deploy, add GPUs in the following way:

```
resources:
  devices:
    - type: nvidia_gpu
      name: {gpu_type}
      count: {gpu_count}
```
---------
"""

LEGACY_GPU_COUNT_WARNING_MESSAGE_TEMPLATE = """
---------
The `gpu_count` field is deprecated. Please remove this field from your
deployment Python script or YAML Spec.

If you are using Python SDK, add GPUs in the following way:

```
from servicefoundry import NvidiaGPU, Resources
...

  resources=Resources(
    ...
    devices=[NvidiaGPU(count={gpu_count})],
  )
```

If you are using YAML Spec to deploy, add GPUs in the following way:

```
resources:
  devices:
    - type: nvidia_gpu
      count: {gpu_count}
```
---------
"""


class CUDAVersion(str, enum.Enum):
    CUDA_11_0_CUDNN8 = "11.0-cudnn8"
    CUDA_11_1_CUDNN8 = "11.1-cudnn8"
    CUDA_11_2_CUDNN8 = "11.2-cudnn8"
    CUDA_11_3_CUDNN8 = "11.3-cudnn8"
    CUDA_11_4_CUDNN8 = "11.4-cudnn8"
    CUDA_11_5_CUDNN8 = "11.5-cudnn8"
    CUDA_11_6_CUDNN8 = "11.6-cudnn8"
    CUDA_11_7_CUDNN8 = "11.7-cudnn8"
    CUDA_11_8_CUDNN8 = "11.8-cudnn8"
    CUDA_12_0_CUDNN8 = "12.0-cudnn8"
    CUDA_12_1_CUDNN8 = "12.1-cudnn8"
    CUDA_12_2_CUDNN8 = "12.2-cudnn8"


class GPUType(str, enum.Enum):
    P4 = "P4"
    P100 = "P100"
    V100 = "V100"
    T4 = "T4"
    A10G = "A10G"
    A100_40GB = "A100_40GB"
    A100_80GB = "A100_80GB"
    L4 = "L4"
    H100_80GB = "H100_80GB"
    H100_94GB = "H100_94GB"


class TPUType(str, enum.Enum):
    V4_PODSLICE = "tpu-v4-podslice"
    V5_LITE_DEVICE = "tpu-v5-lite-device"
    V5_LITE_PODSLICE = "tpu-v5-lite-podslice"
    V5P_SLICE = "tpu-v5p-slice"


class AWSInferentiaAccelerator(str, enum.Enum):
    INF1 = "INF1"
    INF2 = "INF2"


class PatchedModelBase(BaseModel):
    class Config:
        extra = "forbid"


class DockerFileBuild(models.DockerFileBuild, PatchedModelBase):
    type: constr(regex=r"^dockerfile$") = "dockerfile"

    @validator("build_args")
    def validate_build_args(cls, value):
        if not isinstance(value, dict):
            raise TypeError("build_args should be of type dict")
        for k, v in value.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise TypeError("build_args should have keys and values as string")
            if not k.strip() or not v.strip():
                raise ValueError("build_args cannot have empty keys or values")
        return value


class PythonBuild(models.PythonBuild, PatchedModelBase):
    type: constr(regex=r"^tfy-python-buildpack$") = "tfy-python-buildpack"

    @root_validator
    def validate_python_version_when_cuda_version(cls, values):
        if values.get("cuda_version"):
            python_version = values.get("python_version")
            if python_version and not re.match(r"^3\.\d+$", python_version):
                raise ValueError(
                    f'`python_version` must be 3.x (e.g. "3.9") when `cuda_version` field is '
                    f"provided but got {python_version!r}. If you are adding a "
                    f'patch version, please remove it (e.g. "3.9.2" should be "3.9")'
                )
        return values


class RemoteSource(models.RemoteSource, PatchedModelBase):
    type: constr(regex=r"^remote$") = "remote"


class LocalSource(models.LocalSource, PatchedModelBase):
    type: constr(regex=r"^local$") = "local"


class Build(models.Build, PatchedModelBase):
    type: constr(regex=r"^build$") = "build"
    build_source: Union[
        models.RemoteSource, models.GitSource, models.LocalSource
    ] = Field(default_factory=LocalSource)


class Manual(models.Manual, PatchedModelBase):
    type: constr(regex=r"^manual$") = "manual"


class Schedule(models.Schedule, PatchedModelBase):
    type: constr(regex=r"^scheduled$") = "scheduled"


class GitSource(models.GitSource, PatchedModelBase):
    type: constr(regex=r"^git$") = "git"


class HttpProbe(models.HttpProbe, PatchedModelBase):
    type: constr(regex=r"^http$") = "http"


class BasicAuthCreds(models.BasicAuthCreds, PatchedModelBase):
    type: constr(regex=r"^basic_auth$") = "basic_auth"


class HealthProbe(models.HealthProbe, PatchedModelBase):
    pass


class Image(models.Image, PatchedModelBase):
    type: constr(regex=r"^image$") = "image"


class Port(models.Port, PatchedModelBase):
    pass

    @root_validator(pre=True)
    def verify_host(cls, values):
        expose = values.get("expose", True)
        host = values.get("host", None)
        if expose:
            if not host:
                raise ValueError("Host must be provided to expose port")
            if not (
                re.fullmatch(
                    r"^((([a-zA-Z0-9\-]{1,63}\.)([a-zA-Z0-9\-]{1,63}\.)*([A-Za-z]{1,63}))|(((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))$",
                    host,
                )
            ):
                raise ValueError(
                    "Invalid value for `host`. A valid host must contain only alphanumeric letters and hypens e.g.  `ai.example.com`, `app.truefoundry.com`.\nYou can get the list of configured hosts for the cluster from the Integrations > Clusters page. Please see https://docs.truefoundry.com/docs/checking-configured-domain for more information."
                )
        else:
            if host:
                raise ValueError("Cannot pass host when expose=False")

        return values


class Resources(models.Resources, PatchedModelBase):
    @root_validator(pre=False)
    def warn_gpu_count_type_depreciation(cls, values):
        gpu_count = values.get("gpu_count")
        gpu_type = None
        node = values.get("node")
        if node:
            if isinstance(node, NodeSelector):
                gpu_type = node.gpu_type
            elif isinstance(node, Mapping):
                gpu_count = node.get("gpu_type")

        if gpu_count and gpu_type:
            warnings.warn(
                LEGACY_GPU_TYPE_COUNT_WARNING_MESSAGE_TEMPLATE.format(
                    gpu_type=gpu_type,
                    gpu_count=gpu_count,
                ),
                category=FutureWarning,
                stacklevel=2,
            )
        elif gpu_count:
            warnings.warn(
                LEGACY_GPU_COUNT_WARNING_MESSAGE_TEMPLATE.format(
                    gpu_count=gpu_count,
                ),
                category=FutureWarning,
                stacklevel=2,
            )
        return values


class Param(models.Param, PatchedModelBase):
    pass


class CPUUtilizationMetric(models.CPUUtilizationMetric, PatchedModelBase):
    type: constr(regex=r"^cpu_utilization$") = "cpu_utilization"


class RPSMetric(models.RPSMetric, PatchedModelBase):
    type: constr(regex=r"^rps$") = "rps"


class CronMetric(models.CronMetric, PatchedModelBase):
    type: constr(regex=r"^cron$") = "cron"


class ServiceAutoscaling(models.ServiceAutoscaling, PatchedModelBase):
    pass


class AsyncServiceAutoscaling(models.AsyncServiceAutoscaling, PatchedModelBase):
    pass


class Autoscaling(ServiceAutoscaling):
    def __init__(self, **kwargs):
        warnings.warn(
            "`servicefoundry.Autoscaling` is deprecated and will be removed in a future version. "
            "Please use `servicefoundry.ServiceAutoscaling` instead. "
            "You can rename `Autoscaling` to `ServiceAutoscaling` in your script.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**kwargs)


class BlueGreen(models.BlueGreen, PatchedModelBase):
    type: constr(regex=r"^blue_green$") = "blue_green"


class Canary(models.Canary, PatchedModelBase):
    type: constr(regex=r"^canary$") = "canary"


class Rolling(models.Rolling, PatchedModelBase):
    type: constr(regex=r"^rolling_update$") = "rolling_update"


class SecretMount(models.SecretMount, PatchedModelBase):
    type: constr(regex=r"^secret$") = "secret"


class StringDataMount(models.StringDataMount, PatchedModelBase):
    type: constr(regex=r"^string$") = "string"


class VolumeMount(models.VolumeMount, PatchedModelBase):
    type: constr(regex=r"^volume$") = "volume"


class NodeSelector(models.NodeSelector, PatchedModelBase):
    type: constr(regex=r"^node_selector$") = "node_selector"
    gpu_type: Optional[Union[GPUType, str]] = None


class NodepoolSelector(models.NodepoolSelector, PatchedModelBase):
    type: constr(regex=r"^nodepool_selector$") = "nodepool_selector"


class Endpoint(models.Endpoint, PatchedModelBase):
    pass


class TruefoundryImageBase(models.TruefoundryImageBase, PatchedModelBase):
    type: constr(regex=r"^truefoundrybase$") = "truefoundrybase"


class TruefoundryImageFull(models.TruefoundryImageFull, PatchedModelBase):
    type: constr(regex=r"^truefoundryfull$") = "truefoundryfull"


class CodeserverImage(models.CodeserverImage, PatchedModelBase):
    type: constr(regex=r"^codeserver$") = "codeserver"


class HelmRepo(models.HelmRepo, PatchedModelBase):
    type: constr(regex=r"^helm-repo$") = "helm-repo"


class OCIRepo(models.OCIRepo, PatchedModelBase):
    type: constr(regex=r"^oci-repo$") = "oci-repo"


class VolumeBrowser(models.VolumeBrowser, PatchedModelBase):
    pass


class WorkerConfig(models.WorkerConfig, PatchedModelBase):
    pass


class SQSInputConfig(models.SQSInputConfig, PatchedModelBase):
    type: constr(regex=r"^sqs$") = "sqs"


class SQSOutputConfig(models.SQSOutputConfig, PatchedModelBase):
    type: constr(regex=r"^sqs$") = "sqs"


class SQSQueueMetricConfig(models.SQSQueueMetricConfig, PatchedModelBase):
    type: constr(regex=r"^sqs$") = "sqs"


class AWSAccessKeyAuth(models.AWSAccessKeyAuth, PatchedModelBase):
    pass


class NATSInputConfig(models.NATSInputConfig, PatchedModelBase):
    type: constr(regex=r"^nats$") = "nats"


class CoreNATSOutputConfig(models.CoreNATSOutputConfig, PatchedModelBase):
    type: constr(regex=r"^core-nats$") = "core-nats"


class NATSUserPasswordAuth(models.NATSUserPasswordAuth, PatchedModelBase):
    pass


class NATSOutputConfig(models.NATSOutputConfig, PatchedModelBase):
    type: constr(regex=r"^nats$") = "nats"


class NATSMetricConfig(models.NATSMetricConfig, PatchedModelBase):
    type: constr(regex=r"^nats$") = "nats"


class KafkaInputConfig(models.KafkaInputConfig, PatchedModelBase):
    type: constr(regex=r"^kafka$") = "kafka"


class KafkaOutputConfig(models.KafkaOutputConfig, PatchedModelBase):
    type: constr(regex=r"^kafka$") = "kafka"


class KafkaMetricConfig(models.KafkaMetricConfig, PatchedModelBase):
    type: constr(regex=r"^kafka$") = "kafka"


class KafkaSASLAuth(models.KafkaSASLAuth, PatchedModelBase):
    pass


class AMQPInputConfig(models.AMQPInputConfig, PatchedModelBase):
    type: constr(regex=r"^amqp$") = "amqp"


class AMQPOutputConfig(models.AMQPOutputConfig, PatchedModelBase):
    type: constr(regex=r"^amqp$") = "amqp"


class AMQPMetricConfig(models.AMQPMetricConfig, PatchedModelBase):
    type: constr(regex=r"^amqp$") = "amqp"


class AsyncProcessorSidecar(models.AsyncProcessorSidecar, PatchedModelBase):
    pass


class ArtifactsCacheVolume(models.ArtifactsCacheVolume, PatchedModelBase):
    pass


class HuggingfaceArtifactSource(models.HuggingfaceArtifactSource, PatchedModelBase):
    type: constr(regex=r"^huggingface-hub$") = "huggingface-hub"


class TruefoundryArtifactSource(models.TruefoundryArtifactSource, PatchedModelBase):
    type: constr(regex=r"^truefoundry-artifact$") = "truefoundry-artifact"


class ArtifactsDownload(models.ArtifactsDownload, PatchedModelBase):
    pass


class CustomNotebookImage(models.CustomNotebookImage, PatchedModelBase):
    type: constr(regex=r"^customnotebook$") = "customnotebook"


class NvidiaGPU(models.NvidiaGPU, PatchedModelBase):
    type: constr(regex=r"^nvidia_gpu$") = "nvidia_gpu"
    name: Optional[Union[GPUType, str]] = None


class NvidiaMIGGPU(models.NvidiaMIGGPU, PatchedModelBase):
    type: constr(regex=r"^nvidia_mig_gpu$") = "nvidia_mig_gpu"


class NvidiaTimeslicingGPU(models.NvidiaTimeslicingGPU, PatchedModelBase):
    type: constr(regex=r"^nvidia_timeslicing_gpu$") = "nvidia_timeslicing_gpu"


class AWSInferentia(models.AWSInferentia, PatchedModelBase):
    type: constr(regex=r"^aws_inferentia$") = "aws_inferentia"
    name: Optional[Union[AWSInferentiaAccelerator, str]] = None


class GcpTPU(models.GcpTPU, PatchedModelBase):
    type: constr(regex=r"^gcp_tpu$") = "gcp_tpu"
    name: Union[TPUType, constr(regex=r"^tpu-[a-z\d\-]+$")]


class CustomCodeserverImage(models.CustomCodeserverImage, PatchedModelBase):
    type: constr(regex=r"^customcodeserver$") = "customcodeserver"


class CustomSSHServerImage(models.CustomSSHServerImage, PatchedModelBase):
    type: constr(regex=r"^custom-ssh-server$") = "custom-ssh-server"


class SSHServerImage(models.SSHServerImage, PatchedModelBase):
    type: constr(regex=r"^ssh-server") = "ssh-server"


class TruefoundryImageCuda1180(models.TruefoundryImageCuda1180, PatchedModelBase):
    type: constr(regex=r"^truefoundrycuda1180$") = "truefoundrycuda1180"


class TruefoundryImageCuda1211(models.TruefoundryImageCuda1211, PatchedModelBase):
    type: constr(regex=r"^truefoundrycuda1211$") = "truefoundrycuda1211"


class DynamicVolumeConfig(models.DynamicVolumeConfig, PatchedModelBase):
    type: constr(regex=r"^dynamic$") = "dynamic"


class StaticVolumeConfig(models.StaticVolumeConfig, PatchedModelBase):
    type: constr(regex=r"^static$") = "static"



import json
from typing import Dict, Optional, Sequence

import rich_click as click
from click import ClickException

from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application


@click.group(name="trigger", cls=GROUP_CLS)
def trigger_command():
    """
    Trigger a Job asynchronously
    """
    pass


@click.command(
    name="job",
    cls=COMMAND_CLS,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    required=True,
    help="FQN of the deployment of the Job. This can be found on the Job details page.",
)
@click.option("--command", type=click.STRING, required=False, help="Command to run")
@click.argument(
    "params",
    type=click.STRING,
    nargs=-1,
    required=False,
)
@handle_exception_wrapper
def trigger_job(application_fqn: str, command: Optional[Sequence[str]], params):
    """
    Trigger a Job on Truefoundry asynchronously

        [b]tfy trigger job --application-fqn "my-cluster:my-workspace:my-job"[/]

    \n
    Additionally, you can either pass `--command` or params (if defined in the spec)\n


    Passing a command:

        [b]tfy trigger job --application-fqn "my-cluster:my-workspace:my-job" --command "python run.py"[/]
    \n

    Passing params:

        [b]tfy trigger job --application-fqn "my-cluster:my-workspace:my-job" -- --param1_name param1_value --param2_name param2_value ...[/]
    """
    if params:
        params_dict = {}
        if len(params) % 2 != 0:
            raise ClickException(
                f"Found odd number of argument pairs: {params}. "
                "Perhaps you forgot to pass a value for one of the params? "
                "Job params should be passed in the "
                "format `--param1_name param1_value --param2_name param2_value ...`"
            )
        for i in range(0, len(params), 2):
            key = params[i]
            value = params[i + 1]
            if not key.startswith("--"):
                raise ClickException(
                    f"Got ambiguous argument {key!r} in params: {params}. "
                    f"Param names should be prefixed with '--' i.e. "
                    "Job params should be passed in the "
                    "format `--param1_name param1_value --param2_name param2_value ...`"
                )
            key = key.lstrip("-")
            params_dict[key] = value

    application.trigger_job(
        application_fqn=application_fqn, command=command, params=params
    )


def get_trigger_command():
    trigger_command.add_command(trigger_job)
    return trigger_command



from servicefoundry.auto_gen.models import (
    AppProtocol,
    CanaryStep,
    CapacityType,
    ConcurrencyPolicy,
    Kustomize,
    ParamType,
    Protocol,
)
from servicefoundry.core import login, logout
from servicefoundry.lib.dao.application import (
    delete_application,
    get_application,
    get_job_run,
    list_applications,
    list_job_runs,
    terminate_job_run,
    trigger_job,
)
from servicefoundry.lib.dao.version import (
    get_version as get_application_version,
    list_versions as list_application_versions,
)
from servicefoundry.lib.dao.workspace import (
    delete_workspace,
    get_workspace_by_fqn,
    list_workspaces,
)
from servicefoundry.v2.lib.deployable_patched_models import (
    Application,
    AsyncService,
    Codeserver,
    Helm,
    Job,
    Notebook,
    Service,
    SSHServer,
    Volume,
)

from servicefoundry.python_deploy_generator import convert_deployment_config_to_python
from servicefoundry.v2.lib.patched_models import (
    AMQPInputConfig,
    AMQPMetricConfig,
    AMQPOutputConfig,
    ArtifactsCacheVolume,
    ArtifactsDownload,
    AsyncProcessorSidecar,
    AsyncServiceAutoscaling,
    Autoscaling,
    AWSAccessKeyAuth,
    AWSInferentia,
    AWSInferentiaAccelerator,
    BasicAuthCreds,
    BlueGreen,
    Build,
    Canary,
    CodeserverImage,
    CoreNATSOutputConfig,
    CPUUtilizationMetric,
    CronMetric,
    CUDAVersion,
    CustomCodeserverImage,
    CustomNotebookImage,
    CustomSSHServerImage,
    DockerFileBuild,
    DynamicVolumeConfig,
    Endpoint,
    GcpTPU,
    GitSource,
    GPUType,
    HealthProbe,
    HelmRepo,
    HttpProbe,
    HuggingfaceArtifactSource,
    Image,
    KafkaInputConfig,
    KafkaMetricConfig,
    KafkaOutputConfig,
    KafkaSASLAuth,
    LocalSource,
    Manual,
    NATSInputConfig,
    NATSMetricConfig,
    NATSOutputConfig,
    NATSUserPasswordAuth,
    NodepoolSelector,
    NodeSelector,
    NvidiaGPU,
    NvidiaMIGGPU,
    NvidiaTimeslicingGPU,
    OCIRepo,
    Param,
    Port,
    PythonBuild,
    RemoteSource,
    Resources,
    Rolling,
    RPSMetric,
    Schedule,
    SecretMount,
    ServiceAutoscaling,
    SQSInputConfig,
    SQSOutputConfig,
    SQSQueueMetricConfig,
    SSHServerImage,
    StaticVolumeConfig,
    StringDataMount,
    TPUType,
    TruefoundryArtifactSource,
    TruefoundryImageBase,
    TruefoundryImageCuda1180,
    TruefoundryImageCuda1211,
    TruefoundryImageFull,
    VolumeBrowser,
    VolumeMount,
    WorkerConfig,
)
from servicefoundry.version import __version__



import concurrent.futures
import math
from typing import Dict, List, Optional

import tqdm
from langchain.embeddings.base import Embeddings
from langchain.pydantic_v1 import BaseModel, Extra, Field, root_validator

from servicefoundry.langchain.utils import (
    ModelParameters,
    requests_retry_session,
    validate_tfy_environment,
)
from servicefoundry.logger import logger

EMBEDDER_BATCH_SIZE = 32
PARALLEL_WORKERS = 4


class TrueFoundryEmbeddings(BaseModel, Embeddings):
    """`TrueFoundry LLM Gateway` embedding models API.

    To use, you must have the environment variable ``TFY_API_KEY`` set with your API key and ``TFY_HOST`` set with your host or pass it
    as a named parameter to the constructor.
    """

    model: str = Field(description="The model to use for embedding.")
    """The model to use for embedding."""
    tfy_llm_gateway_url: Optional[str] = Field(default=None)
    """TrueFoundry LLM Gateway endpoint URL. Automatically inferred from env var `TFY_LLM_GATEWAY_URL` if not provided."""
    tfy_api_key: Optional[str] = Field(default=None)
    """TrueFoundry API Key. Automatically inferred from env var `TFY_API_KEY` if not provided."""
    model_parameters: Optional[dict] = Field(default_factory=dict)
    """Model parameters"""
    request_timeout: int = Field(default=30)
    """The timeout for the request in seconds."""
    max_retries: int = Field(default=5)
    """The number of retries for HTTP requests."""
    retry_backoff_factor: float = Field(default=0.3)
    """The backoff factor for exponential backoff during retries."""
    batch_size: int = Field(default=EMBEDDER_BATCH_SIZE)
    """The batch size to use for embedding."""
    parallel_workers: int = Field(default=PARALLEL_WORKERS)
    """The number of parallel workers to use for embedding."""

    __private_attributes__ = {"_executor"}

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        allow_population_by_field_name = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        values = validate_tfy_environment(values)
        if not values["tfy_api_key"]:
            raise ValueError(
                f"Did not find `tfy_api_key`, please add an environment variable"
                f" `TFY_API_KEY` which contains it, or pass"
                f"  `tfy_api_key` as a named parameter."
            )
        if not values["tfy_llm_gateway_url"]:
            raise ValueError(
                f"Did not find `tfy_llm_gateway_url`, please add an environment variable"
                f" `TFY_LLM_GATEWAY_URL` which contains it, or pass"
                f"  `tfy_llm_gateway_url` as a named parameter."
            )
        return values

    def _init_private_attributes(self):
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.parallel_workers
        )

    @property
    def _llm_type(self) -> str:
        """Return type of embedding model."""
        return "truefoundry-embeddings"

    def __del__(self):
        """
        Destructor method to clean up the executor when the object is deleted.

        Args:
            None

        Returns:
            None
        """
        self._executor.shutdown()

    def _remote_embed(self, texts, query_mode=False):
        """
        Perform remote embedding using a HTTP POST request to a designated endpoint.

        Args:
            texts (List[str]): A list of text strings to be embedded.
            query_mode (bool): A flag to indicate if running in query mode or in embed mode (indexing).
        Returns:
            List[List[float]]: A list of embedded representations of the input texts.
        """
        session = requests_retry_session(
            retries=self.max_retries, backoff_factor=self.retry_backoff_factor
        )

        payload = {
            "input": texts,
            "model": self.model,
        }

        url = f"{self.tfy_llm_gateway_url}/openai/embeddings"
        logger.debug(
            f"Embedding using - model: {self.model} at endpoint: {url}, for {len(texts)} texts"
        )
        response = session.post(
            url=url,
            json=payload,
            headers={
                "Authorization": f"Bearer {self.tfy_api_key}",
            },
            timeout=self.request_timeout,
        )
        response.raise_for_status()
        output = response.json()
        return [data["embedding"] for data in output["data"]]

    def _embed(self, texts: List[str], query_mode: bool):
        """
        Perform embedding on a list of texts using remote embedding in chunks.

        Args:
            texts (List[str]): A list of text strings to be embedded.
            query_mode (bool): A flag to indicate if running in query mode or in embed mode (indexing).
        Returns:
            List[List[float]]: A list of embedded representations of the input texts.
        """
        embeddings = []

        def _feeder():
            for i in range(0, len(texts), self.batch_size):
                chunk = texts[i : i + self.batch_size]
                yield chunk

        embeddings = list(
            tqdm.tqdm(
                self._executor.map(self._remote_embed, _feeder()),
                total=int(math.ceil(len(texts) / self.batch_size)),
            )
        )
        return [item for batch in embeddings for item in batch]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        Embed a list of text documents.

        Args:
            texts (List[str]): A list of text documents to be embedded.

        Returns:
            List[List[float]]: A list of embedded representations of the input documents.
        """
        return self._embed(texts, query_mode=False)

    def embed_query(self, text: str) -> List[float]:
        """
        Embed a query text.

        Args:
            text (str): The query text to be embedded.

        Returns:
            List[float]: The embedded representation of the input query text.
        """
        return self._embed([text], query_mode=True)[0]



import ast
import io
import re
from typing import List

from rich.console import Console
from rich.pretty import pprint

from servicefoundry import Application


def generate_code(
    symbols_to_import: List[str],
    application_type: str,
    spec_repr: str,
    workspace_fqn: str,
):
    symbols = ",".join(symbols_to_import)
    code = f"""\
import logging
from truefoundry.deploy import (
    {symbols},
)
logging.basicConfig(level=logging.INFO)

{application_type} = {spec_repr}

{application_type}.deploy(workspace_fqn="{workspace_fqn}")\
"""
    return code


def extract_class_names(code):
    tree = ast.parse(code)

    # Function to extract keywords from the AST
    def extract_class_names_from_ast_tree(node):
        keywords = set()
        for child_node in ast.iter_child_nodes(node):
            if isinstance(child_node, ast.Call):
                keywords.add(child_node.func.id)
            keywords.update(extract_class_names_from_ast_tree(child_node))
        return keywords

    # Get keywords from the main body of the code
    main_keywords = extract_class_names_from_ast_tree(tree)
    return list(main_keywords)


def replace_enums_with_values(raw_str):
    # required to replace enums of format <AppProtocol.HTTP: 'http'> with 'http'
    pattern = r'<([a-zA-Z0-9_]+).[a-zA-Z0-9_]+: [\'"](.+)[\'"]>'
    replacement = r"'\2'"

    result = re.sub(pattern, replacement, raw_str)
    return result


def remove_none_type_fields(code):
    lines = code.split("\n")
    new_lines = [
        line
        for line in lines
        if not (line.endswith("=None") or line.endswith("=None,"))
    ]
    formatted_code = "\n".join(new_lines)
    return formatted_code


def remove_type_field(code):
    lines = code.split("\n")
    new_lines = [re.sub(r'^[ \t]*type=[\'"][^"]*[\'"],', "", line) for line in lines]
    return "\n".join(new_lines)


def add_deploy_line(code, workspace_fqn, application_type):
    deploy_line = f"{application_type}.deploy('workspace_fqn={workspace_fqn}')"
    return code + "\n" + deploy_line


def get_python_repr(obj):
    stream = io.StringIO()
    console = Console(file=stream, no_color=True, highlighter=None)
    pprint(obj, expand_all=True, console=console, indent_guides=False)
    return stream.getvalue()


COMMENT_FOR_LOCAL_SOURCE = """# Set build_source=LocalSource(local_build=False), in order to deploy code from your local.
# With local_build=False flag, docker image will be built on cloud instead of local
# Else it will try to use docker installed on your local machine to build the image"""


def add_local_source_comment(code):
    lines = code.split("\n")
    new_lines = []
    for line in lines:
        if line.lstrip(" ").startswith("build_source=GitSource"):
            new_lines.append(COMMENT_FOR_LOCAL_SOURCE)
        new_lines.append(line)
    return "\n".join(new_lines)


def convert_deployment_config_to_python(workspace_fqn: str, deployment_config: dict):
    """
    Convert a deployment config to a python file that can be used to deploy to a workspace
    """
    application = Application.parse_obj(deployment_config)
    application_type = application.__root__.type

    spec_repr = get_python_repr(application.__root__)
    spec_repr = replace_enums_with_values(spec_repr)
    spec_repr = remove_none_type_fields(spec_repr)
    spec_repr = remove_type_field(spec_repr)

    # extract class names to import
    symbols_to_import = extract_class_names(spec_repr)

    # check if GitSource exists in array of symbols to import
    if "GitSource" in symbols_to_import:
        symbols_to_import.append("LocalSource")

    generated_code = generate_code(
        symbols_to_import=symbols_to_import,
        application_type=application_type,
        spec_repr=spec_repr,
        workspace_fqn=workspace_fqn,
    )

    if "GitSource" in symbols_to_import:
        generated_code = add_local_source_comment(generated_code)

    return generated_code



import contextlib
import sys
from enum import Enum

import rich_click as click
import yaml
from rich.pretty import pprint

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_entity_obj, print_json
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import application as application_lib
from servicefoundry.lib.dao import version as version_lib
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.lib.util import is_experimental_env_set

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="get", cls=GROUP_CLS)
def get_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Get Truefoundry resources

    \b
    Supported resources:
    - Workspace
    - Application
    - Application Version
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Get Workspace details")
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the Workspace",
    required=True,
)
@handle_exception_wrapper
def get_workspace(workspace_fqn):
    workspace = workspace_lib.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if CliConfig.get("json"):
        print_json(data=workspace.dict())
    else:
        print_entity_obj("Workspace", workspace)


@click.command(name="application", cls=COMMAND_CLS, help="Get Application details")
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@handle_exception_wrapper
def get_application(application_fqn):
    application = application_lib.get_application(application_fqn=application_fqn)
    if CliConfig.get("json"):
        print_json(data=application.dict())
    else:
        print_entity_obj(
            "Application",
            application,
        )


@click.command(
    name="application-version", cls=COMMAND_CLS, help="Get Application Version details"
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@click.option(
    "--version",
    type=click.STRING,
    default=None,
    help="Version number of the application deployment",
    required=True,
)
@handle_exception_wrapper
def get_version(application_fqn, version):
    version = version_lib.get_version(application_fqn=application_fqn, version=version)
    if CliConfig.get("json"):
        print_json(data=version.dict())
    else:
        print_entity_obj("Version", version)


@click.command(
    name="spec", cls=COMMAND_CLS, help="Get YAML/Python Spec for an Application Version"
)
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@click.option(
    "--version",
    type=click.STRING,
    default=None,
    help="Version number of the application deployment",
    required=True,
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(
        [
            "yml",
            "yaml",
            "json",
            "py",
            "python",
        ]
    ),
    default="yaml",
    help="Output format for the spec",
    required=False,
)
@handle_exception_wrapper
def get_spec(application_fqn, version, output):
    version = version_lib.get_version(application_fqn=application_fqn, version=version)
    if output in ["yml", "yaml"]:
        yaml.safe_dump(version.manifest.dict(), sys.stdout, indent=2)
    elif output in ["json"]:
        print_json(version.manifest.dict())
    elif output in ["py", "python"]:
        from servicefoundry.v2.lib.deployable_patched_models import Application

        manifest = version.manifest.dict()
        instance = Application.parse_obj(manifest).__root__

        # TODO (chiragjn): Can we somehow just enable `use_enum_values` on all Pydantic classes?

        @contextlib.contextmanager
        def _monkey_patch_enum_repr():
            def new_repr(self):
                # return "%r" % (self._value_)
                return "%s.%s" % (self.__class__.__name__, self._name_)

            enum_subclasses = [
                es
                for es in Enum.__subclasses__()
                if es.__module__.startswith("servicefoundry.")
            ]
            original_reprs = []
            for es in enum_subclasses:
                original_reprs.append(es)
                es.__repr__ = new_repr

            yield

            for es, og_repr in zip(enum_subclasses, original_reprs):
                es.__repr__ = og_repr

        with _monkey_patch_enum_repr():
            pprint(
                instance,
                indent_guides=False,
                max_length=88,
                expand_all=True,
                console=console,
            )


@click.command(name="job-run", cls=COMMAND_CLS, help="Get Job Run")
@click.option(
    "--application-fqn",
    "--application_fqn",
    type=click.STRING,
    default=None,
    help="FQN of the application",
    required=True,
)
@click.option(
    "--job-run-name",
    "--job_run_name",
    type=click.STRING,
    default=None,
    help="Run name of the job",
    required=True,
)
@handle_exception_wrapper
def get_job_run(application_fqn, job_run_name):
    job_run = application_lib.get_job_run(
        application_fqn=application_fqn, job_run_name=job_run_name
    )
    if CliConfig.get("json"):
        print_json(data=job_run.dict())
    else:
        print_entity_obj("Job Run", job_run)


def get_get_command():
    get_command.add_command(get_workspace)
    get_command.add_command(get_application)
    get_command.add_command(get_version)
    get_command.add_command(get_job_run)
    if is_experimental_env_set():
        get_command.add_command(get_spec)
    return get_command



try:
    import langchain
except Exception as ex:
    raise Exception(
        "Failed to import langchain."
        " Please install langchain by using `pip install langchain` command"
    ) from ex
from servicefoundry.langchain.deprecated import TruefoundryLLM, TruefoundryPlaygroundLLM
from servicefoundry.langchain.truefoundry_chat import TrueFoundryChat
from servicefoundry.langchain.truefoundry_embeddings import TrueFoundryEmbeddings
from servicefoundry.langchain.truefoundry_llm import TrueFoundryLLM
from servicefoundry.langchain.utils import ModelParameters
